#include "atlas_misc.h"
#ifndef PFYDIST
   #define PFYDIST 16
#endif
#ifndef PFXDIST
   #define PFXDIST 0
#endif
#ifndef PFADIST
   #define PFADIST 0
#endif

#if !defined(ATL_3DNow) && !defined(ATL_SSE1) && \
    (PFADIST != 0 || PFXDIST != 0 || PFYDIST != 0)
   #include "atlas_prefetch.h"
#endif

#if defined(ATL_3DNow) || defined(ATL_SSE1)
   #ifndef PFIY
      #define PFIY prefetchnta
   #endif
   #ifndef PFIX
      #define PFIX prefetchnta
   #endif
   #ifndef PFIA
      #ifdef ATL_3DNow
         #define PFIA prefetchw
      #else
         #define PFIA prefetcht0
      #endif
   #endif
#endif
/*
 * X & A are prefetched in M loop PF[A,X]DIST (in bytes) ahead
 */
#if PFADIST == 0                /* flag for no prefetch */
   #define prefA(mem)
#else
   #if defined(ATL_3DNow) || defined(ATL_SSE1)
      #define prefA(mem) __asm__ __volatile__ \
         (Mstr(PFIA) " %0" : : "m" (*(((char *)(mem))+PFADIST)))
   #else
      #if PFLVL == 2
         #define prefA(mem) ATL_pfl2W(((char*)mem)+PFADIST)
      #else
         #define prefA(mem) ATL_pfl1W(((char*)mem)+PFADIST)
      #endif
   #endif
#endif
#if PFXDIST == 0                /* flag for no prefetch */
   #define prefX(mem)
#else
   #if defined(ATL_3DNow) || defined(ATL_SSE1)
      #define prefX(mem) __asm__ __volatile__ \
         (Mstr(PFIX) " %0" : : "m" (*(((char *)(mem))+PFXDIST)))
   #else
      #if PFLVL == 2
         #define prefX(mem) ATL_pfl2R(((char*)mem)+PFXDIST)
      #else
         #define prefX(mem) ATL_pfl1R(((char*)mem)+PFXDIST)
      #endif
   #endif
#endif
/*
 * Y is prefetched in N-loop, and always fetches next NU piece
 */
#if PFYDIST == 0                /* flag for no prefetch */
   #define prefY(mem)
#else
   #if defined(ATL_3DNow) || defined(ATL_SSE1)
      #define prefY(mem) \
         __asm__ __volatile__ (Mstr(PFIY) " %0" : : "m" (*(((char *)(mem)))))
   #else
      #if PFLVL == 2
         #define prefY(mem) ATL_pfl2R(mem)
      #else
         #define prefY(mem) ATL_pfl1R(mem)
      #endif
   #endif
#endif
#ifndef ATL_CINT
   #define ATL_CINT const int
#endif
#ifndef ATL_INT
   #define ATL_INT int
#endif
void ATL_UGERK(ATL_CINT M, ATL_CINT N, const float *X, const float *Y, 
               float *A, ATL_CINT lda)
/*
 * 32x16 unrolled r1_c.
 * Extracted from ATLAS/tune/blas/gemv/atlas-l2g.base
 */
{
   ATL_CINT N16=(N/16)*16, M32=(M/32)*32, lda16=lda*16;
   ATL_INT j;

   for (j=N16; j; j -= 16, A += lda16, Y += 16)
   {
      float *A0=A, *A1=A0+lda, *A2=A1+lda, *A3=A2+lda, *A4=A3+lda, *A5=A4+lda,
            *A6=A5+lda, *A7=A6+lda, *A8=A7+lda, *A9=A8+lda, *A10=A9+lda,
            *A11=A10+lda, *A12=A11+lda, *A13=A12+lda, *A14=A13+lda,
            *A15=A14+lda;
      const register float y0=Y[0], y1=Y[1], y2=Y[2], y3=Y[3], y4=Y[4], y5=Y[5],
                           y6=Y[6], y7=Y[7], y8=Y[8], y9=Y[9], y10=Y[10],
                           y11=Y[11], y12=Y[12], y13=Y[13], y14=Y[14],
                           y15=Y[15];
      ATL_INT i;
      prefY(Y+16+16-1);
      for (i=0; i < M32; i += 32)
      {
         const register float x0=X[i+0], x1=X[i+1], x2=X[i+2], x3=X[i+3],
                              x4=X[i+4], x5=X[i+5], x6=X[i+6], x7=X[i+7],
                              x8=X[i+8], x9=X[i+9], x10=X[i+10], x11=X[i+11],
                              x12=X[i+12], x13=X[i+13], x14=X[i+14],
                              x15=X[i+15], x16=X[i+16], x17=X[i+17],
                              x18=X[i+18], x19=X[i+19], x20=X[i+20],
                              x21=X[i+21], x22=X[i+22], x23=X[i+23],
                              x24=X[i+24], x25=X[i+25], x26=X[i+26],
                              x27=X[i+27], x28=X[i+28], x29=X[i+29],
                              x30=X[i+30], x31=X[i+31];

         prefX(X);
         A0[i+0] += y0 * x0;
         prefA(A0);
         A1[i+0] += y1 * x0;
         prefA(A1);
         A2[i+0] += y2 * x0;
         prefA(A2);
         A3[i+0] += y3 * x0;
         prefA(A3);
         A4[i+0] += y4 * x0;
         prefA(A4);
         A5[i+0] += y5 * x0;
         prefA(A5);
         A6[i+0] += y6 * x0;
         prefA(A6);
         A7[i+0] += y7 * x0;
         prefA(A7);
         A8[i+0] += y8 * x0;
         prefA(A8);
         A9[i+0] += y9 * x0;
         prefA(A9);
         A10[i+0] += y10 * x0;
         prefA(A10);
         A11[i+0] += y11 * x0;
         prefA(A11);
         A12[i+0] += y12 * x0;
         prefA(A12);
         A13[i+0] += y13 * x0;
         prefA(A13);
         A14[i+0] += y14 * x0;
         prefA(A14);
         A15[i+0] += y15 * x0;
         prefA(A15);
         A0[i+1] += y0 * x1;
         A1[i+1] += y1 * x1;
         A2[i+1] += y2 * x1;
         A3[i+1] += y3 * x1;
         A4[i+1] += y4 * x1;
         A5[i+1] += y5 * x1;
         A6[i+1] += y6 * x1;
         A7[i+1] += y7 * x1;
         A8[i+1] += y8 * x1;
         A9[i+1] += y9 * x1;
         A10[i+1] += y10 * x1;
         A11[i+1] += y11 * x1;
         A12[i+1] += y12 * x1;
         A13[i+1] += y13 * x1;
         A14[i+1] += y14 * x1;
         A15[i+1] += y15 * x1;
         A0[i+2] += y0 * x2;
         A1[i+2] += y1 * x2;
         A2[i+2] += y2 * x2;
         A3[i+2] += y3 * x2;
         A4[i+2] += y4 * x2;
         A5[i+2] += y5 * x2;
         A6[i+2] += y6 * x2;
         A7[i+2] += y7 * x2;
         A8[i+2] += y8 * x2;
         A9[i+2] += y9 * x2;
         A10[i+2] += y10 * x2;
         A11[i+2] += y11 * x2;
         A12[i+2] += y12 * x2;
         A13[i+2] += y13 * x2;
         A14[i+2] += y14 * x2;
         A15[i+2] += y15 * x2;
         A0[i+3] += y0 * x3;
         A1[i+3] += y1 * x3;
         A2[i+3] += y2 * x3;
         A3[i+3] += y3 * x3;
         A4[i+3] += y4 * x3;
         A5[i+3] += y5 * x3;
         A6[i+3] += y6 * x3;
         A7[i+3] += y7 * x3;
         A8[i+3] += y8 * x3;
         A9[i+3] += y9 * x3;
         A10[i+3] += y10 * x3;
         A11[i+3] += y11 * x3;
         A12[i+3] += y12 * x3;
         A13[i+3] += y13 * x3;
         A14[i+3] += y14 * x3;
         A15[i+3] += y15 * x3;
         A0[i+4] += y0 * x4;
         A1[i+4] += y1 * x4;
         A2[i+4] += y2 * x4;
         A3[i+4] += y3 * x4;
         A4[i+4] += y4 * x4;
         A5[i+4] += y5 * x4;
         A6[i+4] += y6 * x4;
         A7[i+4] += y7 * x4;
         A8[i+4] += y8 * x4;
         A9[i+4] += y9 * x4;
         A10[i+4] += y10 * x4;
         A11[i+4] += y11 * x4;
         A12[i+4] += y12 * x4;
         A13[i+4] += y13 * x4;
         A14[i+4] += y14 * x4;
         A15[i+4] += y15 * x4;
         A0[i+5] += y0 * x5;
         A1[i+5] += y1 * x5;
         A2[i+5] += y2 * x5;
         A3[i+5] += y3 * x5;
         A4[i+5] += y4 * x5;
         A5[i+5] += y5 * x5;
         A6[i+5] += y6 * x5;
         A7[i+5] += y7 * x5;
         A8[i+5] += y8 * x5;
         A9[i+5] += y9 * x5;
         A10[i+5] += y10 * x5;
         A11[i+5] += y11 * x5;
         A12[i+5] += y12 * x5;
         A13[i+5] += y13 * x5;
         A14[i+5] += y14 * x5;
         A15[i+5] += y15 * x5;
         A0[i+6] += y0 * x6;
         A1[i+6] += y1 * x6;
         A2[i+6] += y2 * x6;
         A3[i+6] += y3 * x6;
         A4[i+6] += y4 * x6;
         A5[i+6] += y5 * x6;
         A6[i+6] += y6 * x6;
         A7[i+6] += y7 * x6;
         A8[i+6] += y8 * x6;
         A9[i+6] += y9 * x6;
         A10[i+6] += y10 * x6;
         A11[i+6] += y11 * x6;
         A12[i+6] += y12 * x6;
         A13[i+6] += y13 * x6;
         A14[i+6] += y14 * x6;
         A15[i+6] += y15 * x6;
         A0[i+7] += y0 * x7;
         A1[i+7] += y1 * x7;
         A2[i+7] += y2 * x7;
         A3[i+7] += y3 * x7;
         A4[i+7] += y4 * x7;
         A5[i+7] += y5 * x7;
         A6[i+7] += y6 * x7;
         A7[i+7] += y7 * x7;
         A8[i+7] += y8 * x7;
         A9[i+7] += y9 * x7;
         A10[i+7] += y10 * x7;
         A11[i+7] += y11 * x7;
         A12[i+7] += y12 * x7;
         A13[i+7] += y13 * x7;
         A14[i+7] += y14 * x7;
         A15[i+7] += y15 * x7;
         A0[i+8] += y0 * x8;
         A1[i+8] += y1 * x8;
         A2[i+8] += y2 * x8;
         A3[i+8] += y3 * x8;
         A4[i+8] += y4 * x8;
         A5[i+8] += y5 * x8;
         A6[i+8] += y6 * x8;
         A7[i+8] += y7 * x8;
         A8[i+8] += y8 * x8;
         A9[i+8] += y9 * x8;
         A10[i+8] += y10 * x8;
         A11[i+8] += y11 * x8;
         A12[i+8] += y12 * x8;
         A13[i+8] += y13 * x8;
         A14[i+8] += y14 * x8;
         A15[i+8] += y15 * x8;
         A0[i+9] += y0 * x9;
         A1[i+9] += y1 * x9;
         A2[i+9] += y2 * x9;
         A3[i+9] += y3 * x9;
         A4[i+9] += y4 * x9;
         A5[i+9] += y5 * x9;
         A6[i+9] += y6 * x9;
         A7[i+9] += y7 * x9;
         A8[i+9] += y8 * x9;
         A9[i+9] += y9 * x9;
         A10[i+9] += y10 * x9;
         A11[i+9] += y11 * x9;
         A12[i+9] += y12 * x9;
         A13[i+9] += y13 * x9;
         A14[i+9] += y14 * x9;
         A15[i+9] += y15 * x9;
         A0[i+10] += y0 * x10;
         A1[i+10] += y1 * x10;
         A2[i+10] += y2 * x10;
         A3[i+10] += y3 * x10;
         A4[i+10] += y4 * x10;
         A5[i+10] += y5 * x10;
         A6[i+10] += y6 * x10;
         A7[i+10] += y7 * x10;
         A8[i+10] += y8 * x10;
         A9[i+10] += y9 * x10;
         A10[i+10] += y10 * x10;
         A11[i+10] += y11 * x10;
         A12[i+10] += y12 * x10;
         A13[i+10] += y13 * x10;
         A14[i+10] += y14 * x10;
         A15[i+10] += y15 * x10;
         A0[i+11] += y0 * x11;
         A1[i+11] += y1 * x11;
         A2[i+11] += y2 * x11;
         A3[i+11] += y3 * x11;
         A4[i+11] += y4 * x11;
         A5[i+11] += y5 * x11;
         A6[i+11] += y6 * x11;
         A7[i+11] += y7 * x11;
         A8[i+11] += y8 * x11;
         A9[i+11] += y9 * x11;
         A10[i+11] += y10 * x11;
         A11[i+11] += y11 * x11;
         A12[i+11] += y12 * x11;
         A13[i+11] += y13 * x11;
         A14[i+11] += y14 * x11;
         A15[i+11] += y15 * x11;
         A0[i+12] += y0 * x12;
         A1[i+12] += y1 * x12;
         A2[i+12] += y2 * x12;
         A3[i+12] += y3 * x12;
         A4[i+12] += y4 * x12;
         A5[i+12] += y5 * x12;
         A6[i+12] += y6 * x12;
         A7[i+12] += y7 * x12;
         A8[i+12] += y8 * x12;
         A9[i+12] += y9 * x12;
         A10[i+12] += y10 * x12;
         A11[i+12] += y11 * x12;
         A12[i+12] += y12 * x12;
         A13[i+12] += y13 * x12;
         A14[i+12] += y14 * x12;
         A15[i+12] += y15 * x12;
         A0[i+13] += y0 * x13;
         A1[i+13] += y1 * x13;
         A2[i+13] += y2 * x13;
         A3[i+13] += y3 * x13;
         A4[i+13] += y4 * x13;
         A5[i+13] += y5 * x13;
         A6[i+13] += y6 * x13;
         A7[i+13] += y7 * x13;
         A8[i+13] += y8 * x13;
         A9[i+13] += y9 * x13;
         A10[i+13] += y10 * x13;
         A11[i+13] += y11 * x13;
         A12[i+13] += y12 * x13;
         A13[i+13] += y13 * x13;
         A14[i+13] += y14 * x13;
         A15[i+13] += y15 * x13;
         A0[i+14] += y0 * x14;
         A1[i+14] += y1 * x14;
         A2[i+14] += y2 * x14;
         A3[i+14] += y3 * x14;
         A4[i+14] += y4 * x14;
         A5[i+14] += y5 * x14;
         A6[i+14] += y6 * x14;
         A7[i+14] += y7 * x14;
         A8[i+14] += y8 * x14;
         A9[i+14] += y9 * x14;
         A10[i+14] += y10 * x14;
         A11[i+14] += y11 * x14;
         A12[i+14] += y12 * x14;
         A13[i+14] += y13 * x14;
         A14[i+14] += y14 * x14;
         A15[i+14] += y15 * x14;
         A0[i+15] += y0 * x15;
         A1[i+15] += y1 * x15;
         A2[i+15] += y2 * x15;
         A3[i+15] += y3 * x15;
         A4[i+15] += y4 * x15;
         A5[i+15] += y5 * x15;
         A6[i+15] += y6 * x15;
         A7[i+15] += y7 * x15;
         A8[i+15] += y8 * x15;
         A9[i+15] += y9 * x15;
         A10[i+15] += y10 * x15;
         A11[i+15] += y11 * x15;
         A12[i+15] += y12 * x15;
         A13[i+15] += y13 * x15;
         A14[i+15] += y14 * x15;
         A15[i+15] += y15 * x15;
         A0[i+16] += y0 * x16;
         prefA(A0);
         A1[i+16] += y1 * x16;
         prefA(A1);
         A2[i+16] += y2 * x16;
         prefA(A2);
         A3[i+16] += y3 * x16;
         prefA(A3);
         A4[i+16] += y4 * x16;
         prefA(A4);
         A5[i+16] += y5 * x16;
         prefA(A5);
         A6[i+16] += y6 * x16;
         prefA(A6);
         A7[i+16] += y7 * x16;
         prefA(A7);
         A8[i+16] += y8 * x16;
         prefA(A8);
         A9[i+16] += y9 * x16;
         prefA(A9);
         A10[i+16] += y10 * x16;
         prefA(A10);
         A11[i+16] += y11 * x16;
         prefA(A11);
         A12[i+16] += y12 * x16;
         prefA(A12);
         A13[i+16] += y13 * x16;
         prefA(A13);
         A14[i+16] += y14 * x16;
         prefA(A14);
         A15[i+16] += y15 * x16;
         prefA(A15);
         A0[i+17] += y0 * x17;
         A1[i+17] += y1 * x17;
         A2[i+17] += y2 * x17;
         A3[i+17] += y3 * x17;
         A4[i+17] += y4 * x17;
         A5[i+17] += y5 * x17;
         A6[i+17] += y6 * x17;
         A7[i+17] += y7 * x17;
         A8[i+17] += y8 * x17;
         A9[i+17] += y9 * x17;
         A10[i+17] += y10 * x17;
         A11[i+17] += y11 * x17;
         A12[i+17] += y12 * x17;
         A13[i+17] += y13 * x17;
         A14[i+17] += y14 * x17;
         A15[i+17] += y15 * x17;
         A0[i+18] += y0 * x18;
         A1[i+18] += y1 * x18;
         A2[i+18] += y2 * x18;
         A3[i+18] += y3 * x18;
         A4[i+18] += y4 * x18;
         A5[i+18] += y5 * x18;
         A6[i+18] += y6 * x18;
         A7[i+18] += y7 * x18;
         A8[i+18] += y8 * x18;
         A9[i+18] += y9 * x18;
         A10[i+18] += y10 * x18;
         A11[i+18] += y11 * x18;
         A12[i+18] += y12 * x18;
         A13[i+18] += y13 * x18;
         A14[i+18] += y14 * x18;
         A15[i+18] += y15 * x18;
         A0[i+19] += y0 * x19;
         A1[i+19] += y1 * x19;
         A2[i+19] += y2 * x19;
         A3[i+19] += y3 * x19;
         A4[i+19] += y4 * x19;
         A5[i+19] += y5 * x19;
         A6[i+19] += y6 * x19;
         A7[i+19] += y7 * x19;
         A8[i+19] += y8 * x19;
         A9[i+19] += y9 * x19;
         A10[i+19] += y10 * x19;
         A11[i+19] += y11 * x19;
         A12[i+19] += y12 * x19;
         A13[i+19] += y13 * x19;
         A14[i+19] += y14 * x19;
         A15[i+19] += y15 * x19;
         A0[i+20] += y0 * x20;
         A1[i+20] += y1 * x20;
         A2[i+20] += y2 * x20;
         A3[i+20] += y3 * x20;
         A4[i+20] += y4 * x20;
         A5[i+20] += y5 * x20;
         A6[i+20] += y6 * x20;
         A7[i+20] += y7 * x20;
         A8[i+20] += y8 * x20;
         A9[i+20] += y9 * x20;
         A10[i+20] += y10 * x20;
         A11[i+20] += y11 * x20;
         A12[i+20] += y12 * x20;
         A13[i+20] += y13 * x20;
         A14[i+20] += y14 * x20;
         A15[i+20] += y15 * x20;
         A0[i+21] += y0 * x21;
         A1[i+21] += y1 * x21;
         A2[i+21] += y2 * x21;
         A3[i+21] += y3 * x21;
         A4[i+21] += y4 * x21;
         A5[i+21] += y5 * x21;
         A6[i+21] += y6 * x21;
         A7[i+21] += y7 * x21;
         A8[i+21] += y8 * x21;
         A9[i+21] += y9 * x21;
         A10[i+21] += y10 * x21;
         A11[i+21] += y11 * x21;
         A12[i+21] += y12 * x21;
         A13[i+21] += y13 * x21;
         A14[i+21] += y14 * x21;
         A15[i+21] += y15 * x21;
         A0[i+22] += y0 * x22;
         A1[i+22] += y1 * x22;
         A2[i+22] += y2 * x22;
         A3[i+22] += y3 * x22;
         A4[i+22] += y4 * x22;
         A5[i+22] += y5 * x22;
         A6[i+22] += y6 * x22;
         A7[i+22] += y7 * x22;
         A8[i+22] += y8 * x22;
         A9[i+22] += y9 * x22;
         A10[i+22] += y10 * x22;
         A11[i+22] += y11 * x22;
         A12[i+22] += y12 * x22;
         A13[i+22] += y13 * x22;
         A14[i+22] += y14 * x22;
         A15[i+22] += y15 * x22;
         A0[i+23] += y0 * x23;
         A1[i+23] += y1 * x23;
         A2[i+23] += y2 * x23;
         A3[i+23] += y3 * x23;
         A4[i+23] += y4 * x23;
         A5[i+23] += y5 * x23;
         A6[i+23] += y6 * x23;
         A7[i+23] += y7 * x23;
         A8[i+23] += y8 * x23;
         A9[i+23] += y9 * x23;
         A10[i+23] += y10 * x23;
         A11[i+23] += y11 * x23;
         A12[i+23] += y12 * x23;
         A13[i+23] += y13 * x23;
         A14[i+23] += y14 * x23;
         A15[i+23] += y15 * x23;
         A0[i+24] += y0 * x24;
         A1[i+24] += y1 * x24;
         A2[i+24] += y2 * x24;
         A3[i+24] += y3 * x24;
         A4[i+24] += y4 * x24;
         A5[i+24] += y5 * x24;
         A6[i+24] += y6 * x24;
         A7[i+24] += y7 * x24;
         A8[i+24] += y8 * x24;
         A9[i+24] += y9 * x24;
         A10[i+24] += y10 * x24;
         A11[i+24] += y11 * x24;
         A12[i+24] += y12 * x24;
         A13[i+24] += y13 * x24;
         A14[i+24] += y14 * x24;
         A15[i+24] += y15 * x24;
         A0[i+25] += y0 * x25;
         A1[i+25] += y1 * x25;
         A2[i+25] += y2 * x25;
         A3[i+25] += y3 * x25;
         A4[i+25] += y4 * x25;
         A5[i+25] += y5 * x25;
         A6[i+25] += y6 * x25;
         A7[i+25] += y7 * x25;
         A8[i+25] += y8 * x25;
         A9[i+25] += y9 * x25;
         A10[i+25] += y10 * x25;
         A11[i+25] += y11 * x25;
         A12[i+25] += y12 * x25;
         A13[i+25] += y13 * x25;
         A14[i+25] += y14 * x25;
         A15[i+25] += y15 * x25;
         A0[i+26] += y0 * x26;
         A1[i+26] += y1 * x26;
         A2[i+26] += y2 * x26;
         A3[i+26] += y3 * x26;
         A4[i+26] += y4 * x26;
         A5[i+26] += y5 * x26;
         A6[i+26] += y6 * x26;
         A7[i+26] += y7 * x26;
         A8[i+26] += y8 * x26;
         A9[i+26] += y9 * x26;
         A10[i+26] += y10 * x26;
         A11[i+26] += y11 * x26;
         A12[i+26] += y12 * x26;
         A13[i+26] += y13 * x26;
         A14[i+26] += y14 * x26;
         A15[i+26] += y15 * x26;
         A0[i+27] += y0 * x27;
         A1[i+27] += y1 * x27;
         A2[i+27] += y2 * x27;
         A3[i+27] += y3 * x27;
         A4[i+27] += y4 * x27;
         A5[i+27] += y5 * x27;
         A6[i+27] += y6 * x27;
         A7[i+27] += y7 * x27;
         A8[i+27] += y8 * x27;
         A9[i+27] += y9 * x27;
         A10[i+27] += y10 * x27;
         A11[i+27] += y11 * x27;
         A12[i+27] += y12 * x27;
         A13[i+27] += y13 * x27;
         A14[i+27] += y14 * x27;
         A15[i+27] += y15 * x27;
         A0[i+28] += y0 * x28;
         A1[i+28] += y1 * x28;
         A2[i+28] += y2 * x28;
         A3[i+28] += y3 * x28;
         A4[i+28] += y4 * x28;
         A5[i+28] += y5 * x28;
         A6[i+28] += y6 * x28;
         A7[i+28] += y7 * x28;
         A8[i+28] += y8 * x28;
         A9[i+28] += y9 * x28;
         A10[i+28] += y10 * x28;
         A11[i+28] += y11 * x28;
         A12[i+28] += y12 * x28;
         A13[i+28] += y13 * x28;
         A14[i+28] += y14 * x28;
         A15[i+28] += y15 * x28;
         A0[i+29] += y0 * x29;
         A1[i+29] += y1 * x29;
         A2[i+29] += y2 * x29;
         A3[i+29] += y3 * x29;
         A4[i+29] += y4 * x29;
         A5[i+29] += y5 * x29;
         A6[i+29] += y6 * x29;
         A7[i+29] += y7 * x29;
         A8[i+29] += y8 * x29;
         A9[i+29] += y9 * x29;
         A10[i+29] += y10 * x29;
         A11[i+29] += y11 * x29;
         A12[i+29] += y12 * x29;
         A13[i+29] += y13 * x29;
         A14[i+29] += y14 * x29;
         A15[i+29] += y15 * x29;
         A0[i+30] += y0 * x30;
         A1[i+30] += y1 * x30;
         A2[i+30] += y2 * x30;
         A3[i+30] += y3 * x30;
         A4[i+30] += y4 * x30;
         A5[i+30] += y5 * x30;
         A6[i+30] += y6 * x30;
         A7[i+30] += y7 * x30;
         A8[i+30] += y8 * x30;
         A9[i+30] += y9 * x30;
         A10[i+30] += y10 * x30;
         A11[i+30] += y11 * x30;
         A12[i+30] += y12 * x30;
         A13[i+30] += y13 * x30;
         A14[i+30] += y14 * x30;
         A15[i+30] += y15 * x30;
         A0[i+31] += y0 * x31;
         A1[i+31] += y1 * x31;
         A2[i+31] += y2 * x31;
         A3[i+31] += y3 * x31;
         A4[i+31] += y4 * x31;
         A5[i+31] += y5 * x31;
         A6[i+31] += y6 * x31;
         A7[i+31] += y7 * x31;
         A8[i+31] += y8 * x31;
         A9[i+31] += y9 * x31;
         A10[i+31] += y10 * x31;
         A11[i+31] += y11 * x31;
         A12[i+31] += y12 * x31;
         A13[i+31] += y13 * x31;
         A14[i+31] += y14 * x31;
         A15[i+31] += y15 * x31;
      }
      for (i=M32; i < M; i++)
      {
         const register float x0 = X[i];
         A0[i] += y0 * x0;
         A1[i] += y1 * x0;
         A2[i] += y2 * x0;
         A3[i] += y3 * x0;
         A4[i] += y4 * x0;
         A5[i] += y5 * x0;
         A6[i] += y6 * x0;
         A7[i] += y7 * x0;
         A8[i] += y8 * x0;
         A9[i] += y9 * x0;
         A10[i] += y10 * x0;
         A11[i] += y11 * x0;
         A12[i] += y12 * x0;
         A13[i] += y13 * x0;
         A14[i] += y14 * x0;
         A15[i] += y15 * x0;
      }
   }
/*
 * Do remaining columns with NU=1 cleanup
 */
   for (j=N-N16; j; j--, A += lda, Y++)
   {
      float *A0=A;
      const register float y0=Y[0];
      ATL_INT i;
      prefY(Y+1+1-1);
      for (i=0; i < M32; i += 32)
      {
         const register float x0=X[i+0], x1=X[i+1], x2=X[i+2], x3=X[i+3],
                              x4=X[i+4], x5=X[i+5], x6=X[i+6], x7=X[i+7],
                              x8=X[i+8], x9=X[i+9], x10=X[i+10], x11=X[i+11],
                              x12=X[i+12], x13=X[i+13], x14=X[i+14],
                              x15=X[i+15], x16=X[i+16], x17=X[i+17],
                              x18=X[i+18], x19=X[i+19], x20=X[i+20],
                              x21=X[i+21], x22=X[i+22], x23=X[i+23],
                              x24=X[i+24], x25=X[i+25], x26=X[i+26],
                              x27=X[i+27], x28=X[i+28], x29=X[i+29],
                              x30=X[i+30], x31=X[i+31];

         prefX(X);
         A0[i+0] += y0 * x0;
         prefA(A0);
         A0[i+1] += y0 * x1;
         A0[i+2] += y0 * x2;
         A0[i+3] += y0 * x3;
         A0[i+4] += y0 * x4;
         A0[i+5] += y0 * x5;
         A0[i+6] += y0 * x6;
         A0[i+7] += y0 * x7;
         A0[i+8] += y0 * x8;
         A0[i+9] += y0 * x9;
         A0[i+10] += y0 * x10;
         A0[i+11] += y0 * x11;
         A0[i+12] += y0 * x12;
         A0[i+13] += y0 * x13;
         A0[i+14] += y0 * x14;
         A0[i+15] += y0 * x15;
         A0[i+16] += y0 * x16;
         prefA(A0);
         A0[i+17] += y0 * x17;
         A0[i+18] += y0 * x18;
         A0[i+19] += y0 * x19;
         A0[i+20] += y0 * x20;
         A0[i+21] += y0 * x21;
         A0[i+22] += y0 * x22;
         A0[i+23] += y0 * x23;
         A0[i+24] += y0 * x24;
         A0[i+25] += y0 * x25;
         A0[i+26] += y0 * x26;
         A0[i+27] += y0 * x27;
         A0[i+28] += y0 * x28;
         A0[i+29] += y0 * x29;
         A0[i+30] += y0 * x30;
         A0[i+31] += y0 * x31;
      }
      for (i=M32; i < M; i++)
      {
         const register float x0 = X[i];
         A0[i] += y0 * x0;
      }
   }
}
